package jp.deadend.noname.dumbgallery

import android.app.Activity
import android.content.Intent
import android.content.pm.ActivityInfo
import android.net.Uri
import android.os.Bundle
import android.os.Handler
import android.os.Looper
import android.view.Menu
import android.view.MenuItem
import androidx.appcompat.app.AppCompatActivity
import androidx.fragment.app.Fragment
import androidx.fragment.app.FragmentActivity
import androidx.viewpager2.adapter.FragmentStateAdapter
import androidx.preference.PreferenceManager
import androidx.viewpager2.widget.ViewPager2
import jp.deadend.noname.dumbgallery.databinding.ActivityDetailsBinding

class DetailsActivity : AppCompatActivity() {
    private lateinit var binding: ActivityDetailsBinding
    private lateinit var mImages: Array<ImageItem>
    private var isSlideshowMode = false
    private var mDelaySlideshow = DEFAULT_DELAY_SLIDESHOW * 1000L
    private val mHandler = Handler(Looper.getMainLooper())
    private val mRunnable: Runnable = object : Runnable {
        override fun run() {
            val adapter = binding.pager.adapter
            adapter?.let {
                val currentPage = binding.pager.currentItem
                val nextPage = if (currentPage == adapter.itemCount - 1) 0 else currentPage + 1
                binding.pager.setCurrentItem(nextPage, true)
            }

            mHandler.postDelayed(this, mDelaySlideshow)
        }
    }
    private lateinit var mSelected: BooleanArray
    private var prevScrollState = ViewPager2.SCROLL_STATE_IDLE
    private var curScrollState = ViewPager2.SCROLL_STATE_IDLE

    override fun onCreate(savedInstanceState: Bundle?) {
        super.onCreate(savedInstanceState)

        supportActionBar?.setDisplayHomeAsUpEnabled(true)

        val extras = intent.extras
        extras ?: finish()

        val images = (extras?.getParcelable(KEY_DIR_URI) as Uri?)?.let {
            val isRecursive = PreferenceManager.getDefaultSharedPreferences(this)
                .getBoolean(getString(R.string.pref_key_recursive), false)
            ImageItem.imagesInDir(contentResolver, it, isRecursive)
        }
        images ?: finish()
        mImages = images!!
        val sortOrder = (extras?.getSerializable(KEY_SORT_ORDER) as ImageItem.SortOrder?)
        val isDescending = extras?.getBoolean(KEY_SORT_IS_DESCENDING)
        if (sortOrder != null && isDescending != null) { mImages.sort(sortOrder, isDescending) }

        mSelected = extras?.getBooleanArray(KEY_SELECTED) ?: BooleanArray(mImages.size) {false}

        val current = extras?.getInt(KEY_CURRENT_NO) ?: 0
        title = mImages[current].name

        binding = ActivityDetailsBinding.inflate(layoutInflater)
        setContentView(binding.root)
        binding.pager.adapter = DetailsPagerAdapter(this)
        binding.pager.setCurrentItem(current, false)
        binding.pager.offscreenPageLimit = 1

        binding.pager.registerOnPageChangeCallback(object : ViewPager2.OnPageChangeCallback() {
            override fun onPageScrollStateChanged(state: Int) {
                val pageCount = mImages.size
                val curPage = binding.pager.currentItem
                if (curPage == pageCount-1 || curPage == 0) {
                    prevScrollState = curScrollState
                    curScrollState = state
                    if (prevScrollState == ViewPager2.SCROLL_STATE_DRAGGING
                        && curScrollState == ViewPager2.SCROLL_STATE_IDLE) {
                        binding.pager.currentItem = if (curPage == 0) pageCount-1 else 0
                    }
                }
            }
            override fun onPageScrolled(position: Int, positionOffset: Float, positionOffsetPixels: Int) {}
            override fun onPageSelected(position: Int) {
                title = mImages[position].name
                invalidateOptionsMenu()
            }
        })

        requestedOrientation = ActivityInfo.SCREEN_ORIENTATION_LOCKED

        mDelaySlideshow = 1000L *
                PreferenceManager.getDefaultSharedPreferences(this)
                .getInt(getString(R.string.pref_key_slideshow_delay), DEFAULT_DELAY_SLIDESHOW)
    }

    override fun onCreateOptionsMenu(menu: Menu): Boolean {
        menuInflater.inflate(R.menu.menu_details, menu)

        return true
    }

    override fun onPrepareOptionsMenu(menu: Menu?): Boolean {
        if (mSelected[binding.pager.currentItem]) {
            menu?.findItem(R.id.action_select)?.setIcon(R.drawable.ic_star_white_24dp)
        } else {
            menu?.findItem(R.id.action_select)?.setIcon(R.drawable.ic_star_border_white_24dp)
        }
        return super.onPrepareOptionsMenu(menu)
    }

    override fun onOptionsItemSelected(item: MenuItem): Boolean {
        when (item.itemId) {
            android.R.id.home -> finish()
            R.id.action_slideshow -> {
                if (!isSlideshowMode) {
                    item.setIcon(R.drawable.ic_pause_white_24dp)
                    item.setTitle(R.string.menu_stop_slideshow)
                    mHandler.postDelayed(mRunnable, mDelaySlideshow)
                } else {
                    item.setIcon(R.drawable.ic_play_arrow_white_24dp)
                    item.setTitle(R.string.menu_slideshow)
                    mHandler.removeCallbacksAndMessages(null)
                }
                isSlideshowMode = !isSlideshowMode
            }
            R.id.action_select -> {
                val current = binding.pager.currentItem
                if (!mSelected[current]) {
                    item.setIcon(R.drawable.ic_star_white_24dp)
                } else {
                    item.setIcon(R.drawable.ic_star_border_white_24dp)
                }
                mSelected[current] = !mSelected[current]
            }
            R.id.action_rotate -> {
                (supportFragmentManager.fragments.firstOrNull {
                    it is ImageViewFragment
                            && it.arguments?.getInt(ImageViewFragment.POSITION) == binding.pager.currentItem
                } as ImageViewFragment?)?.rotateImage()
//                (supportFragmentManager.findFragmentByTag("f${binding.pager.currentItem}") as ImageViewFragment).rotateImage()
            }
        }
        return super.onOptionsItemSelected(item)
    }

    override fun finish() {
        val intent = Intent().apply { putExtra(KEY_SELECTED, mSelected) }
        setResult(Activity.RESULT_OK, intent)

        super.finish()
    }

    private inner class DetailsPagerAdapter(fa: FragmentActivity) : FragmentStateAdapter(fa) {
        override fun createFragment(position: Int): Fragment {
            val fragment = ImageViewFragment.getInstance(position)
            fragment.setAsset(mImages[position])
            return fragment
        }

        override fun getItemCount() = mImages.size
    }

    companion object {
        const val KEY_DIR_URI = "DetailsDirUri"
        const val KEY_SORT_ORDER = "DetailsSortOrder"
        const val KEY_SORT_IS_DESCENDING = "DetailsIsDescending"
        const val KEY_CURRENT_NO = "DetailsCurrentNo"
        const val KEY_SELECTED = "DetailsSelection"
        const val DEFAULT_DELAY_SLIDESHOW = 3
    }
}