package jp.deadend.noname.dumbgallery

import android.content.ContentResolver
import android.graphics.BitmapFactory
import android.net.Uri
import android.os.Parcelable
import android.provider.DocumentsContract
import android.util.Log
import kotlinx.parcelize.Parcelize
import java.lang.Exception
import java.time.Instant
import java.time.ZoneId
import java.time.format.DateTimeFormatter
import java.time.format.FormatStyle

fun Array<ImageItem>.sort(order: ImageItem.SortOrder, isDescending: Boolean) {
    when (order) {
        ImageItem.SortOrder.NAME -> {
            if (!isDescending) {
                this.sortWith(compareBy { it.name })
            } else {
                this.sortWith(compareByDescending { it.name })
            }
        }
        ImageItem.SortOrder.DATE -> {
            if (!isDescending) {
                this.sortWith(compareBy { it.lastModified })
            } else {
                this.sortWith(compareByDescending { it.lastModified })
            }
        }
        ImageItem.SortOrder.RANDOM -> {
            this.shuffle()
        }
    }

}

@Parcelize
class ImageItem(val uri: Uri, val name: String, val lastModified: Long, val mime: String, val size: Long): Parcelable {
    enum class SortOrder { NAME, DATE, RANDOM }

    fun lastModifiedStr(): String = DateTimeFormatter.ofLocalizedDateTime(FormatStyle.SHORT)
                                        .withZone(ZoneId.systemDefault())
                                        .format(Instant.ofEpochMilli(lastModified))

    fun bytesStr() = bytesToStr(size)

    fun getImageSize(resolver: ContentResolver) : Pair<Int, Int> {
        val decodeOptions = BitmapFactory.Options()
        decodeOptions.inJustDecodeBounds = true
        val stream = resolver.openInputStream(uri) ?: return Pair(0, 0)
        BitmapFactory.decodeStream(stream.buffered(), null, decodeOptions)
        stream.close()

        return Pair(decodeOptions.outWidth, decodeOptions.outHeight)
    }

    companion object {
        private const val IMAGE_MIME = "image/"
        private val IMAGE_PROJECTION = arrayOf(
            DocumentsContract.Document.COLUMN_DOCUMENT_ID,
            DocumentsContract.Document.COLUMN_DISPLAY_NAME,
            DocumentsContract.Document.COLUMN_LAST_MODIFIED,
            DocumentsContract.Document.COLUMN_MIME_TYPE,
            DocumentsContract.Document.COLUMN_SIZE,
        )

        private fun bytesToStr(bytes: Long) = when {
            bytes == Long.MIN_VALUE || bytes < 0 -> "N/A"
            bytes < 1024L -> "$bytes B"
            bytes <= 0xfffccccccccccccL shr 40 -> "%.1f KiB".format(bytes.toDouble() / (0x1 shl 10))
            bytes <= 0xfffccccccccccccL shr 30 -> "%.1f MiB".format(bytes.toDouble() / (0x1 shl 20))
            bytes <= 0xfffccccccccccccL shr 20 -> "%.1f GiB".format(bytes.toDouble() / (0x1 shl 30))
            bytes <= 0xfffccccccccccccL shr 10 -> "%.1f TiB".format(bytes.toDouble() / (0x1 shl 40))
            bytes <= 0xfffccccccccccccL -> "%.1f PiB".format((bytes shr 10).toDouble() / (0x1 shl 40))
            else -> "%.1f EiB".format((bytes shr 20).toDouble() / (0x1 shl 40))
        }

        fun imagesInDir(resolver: ContentResolver, dirUri: Uri, isRecursive: Boolean): Array<ImageItem>? {
            val list = mutableListOf<ImageItem>()
            val dirs = mutableListOf<Uri>()
            try {
                DocumentsContract.buildChildDocumentsUriUsingTree(
                    dirUri, DocumentsContract.getTreeDocumentId(dirUri)
                )?.let { rootUri ->
                    dirs.add(rootUri)

                    while (dirs.isNotEmpty()) {
                        val childrenUri = dirs.removeAt(0)
                        resolver.query(
                            childrenUri, IMAGE_PROJECTION, null, null, null
                        ).use { cursor ->
                            if (cursor != null && cursor.count > 0) {
                                cursor.moveToFirst()
                                do {
                                    val mimeIndex =
                                        cursor.getColumnIndex(DocumentsContract.Document.COLUMN_MIME_TYPE)
                                    val mime = cursor.getString(mimeIndex)
                                    val idIndex =
                                        cursor.getColumnIndex(DocumentsContract.Document.COLUMN_DOCUMENT_ID)
                                    val idStr = cursor.getString(idIndex)
                                    if (mime.startsWith(IMAGE_MIME)) {
                                        val nameIndex =
                                            cursor.getColumnIndex(DocumentsContract.Document.COLUMN_DISPLAY_NAME)
                                        val lastModifiedIndex =
                                            cursor.getColumnIndex(DocumentsContract.Document.COLUMN_LAST_MODIFIED)
                                        val sizeIndex =
                                            cursor.getColumnIndex(DocumentsContract.Document.COLUMN_SIZE)
                                        list.add(
                                            ImageItem(
                                                DocumentsContract.buildDocumentUriUsingTree(rootUri, idStr),
                                                cursor.getString(nameIndex),
                                                cursor.getLong(lastModifiedIndex),
                                                mime,
                                                cursor.getLong(sizeIndex)
                                            )
                                        )
                                    }
                                    if (isRecursive) {
                                        if (mime.equals(DocumentsContract.Document.MIME_TYPE_DIR)) {
                                            dirs.add(DocumentsContract.buildChildDocumentsUriUsingTree(rootUri, idStr))
                                        }
                                    }
                                } while (cursor.moveToNext())
                            }
                        }
                    }

                    return list.toTypedArray()
                }
            } catch (se: SecurityException) {
                Log.e("DumbGallery", "open directory failed")
            } catch (e: Exception) {
                e.printStackTrace()
            }

            return null
        }
    }
}