package jp.deadend.noname.dumbgallery

import android.app.Activity
import android.content.Context
import android.content.Intent
import android.graphics.Color
import android.graphics.PorterDuff
import android.graphics.drawable.ColorDrawable
import android.net.Uri
import androidx.appcompat.app.AppCompatActivity
import android.os.Bundle
import android.provider.DocumentsContract
import android.text.method.ScrollingMovementMethod
import android.util.Log
import android.view.LayoutInflater
import android.view.Menu
import android.view.MenuItem
import android.view.View
import android.view.ViewGroup
import android.view.WindowManager
import android.widget.ArrayAdapter
import android.widget.ImageView
import android.widget.PopupWindow
import android.widget.TextView
import android.widget.Toast
import androidx.activity.result.ActivityResult
import androidx.activity.result.contract.ActivityResultContracts
import androidx.activity.result.contract.ActivityResultContracts.StartActivityForResult
import androidx.core.text.bold
import androidx.core.text.buildSpannedString
import androidx.documentfile.provider.DocumentFile
import androidx.preference.PreferenceManager
import com.bumptech.glide.Glide
import jp.deadend.noname.dumbgallery.databinding.ActivityMainBinding
import jp.deadend.noname.dumbgallery.ImageItem.SortOrder

class MainActivity : AppCompatActivity() {
    private lateinit var binding: ActivityMainBinding
    private var mCurrentDirUri: Uri? = null
    private lateinit var mImages: Array<ImageItem>
    private var mSelected = booleanArrayOf()
    private fun countSelected() = mSelected.count {it}

    private var mSortOrder = SortOrder.NAME
    private var isDescending = false

    private var isSelecting = false

    private val startDetailsActivity =
        registerForActivityResult(StartActivityForResult()) { result : ActivityResult ->
            if (result.resultCode == Activity.RESULT_OK) {
                result.data?.getBooleanArrayExtra(DetailsActivity.KEY_SELECTED)?.let {
                    mSelected = it.clone()
                    binding.gridView.invalidateViews()
                }
            }
        }
    private val openDirLauncher =
        registerForActivityResult(ActivityResultContracts.OpenDocumentTree()) { uri ->
            if (uri != null) {
                mCurrentDirUri = uri
                contentResolver.takePersistableUriPermission(uri, Intent.FLAG_GRANT_READ_URI_PERMISSION or Intent.FLAG_GRANT_WRITE_URI_PERMISSION)
                openDir()
            }
        }
    private val copyToDirLauncher =
        registerForActivityResult(ActivityResultContracts.OpenDocumentTree()) { uri ->
            uri?.let {
                if (copySelectedTo(it)) {
                    Toast.makeText(
                        this,
                        getString(R.string.message_copied, countSelected(), uri.path),
                        Toast.LENGTH_LONG
                    ).show()
                } else {
                    Toast.makeText(
                        this,
                        getString(R.string.error_copy_failed, uri.path),
                        Toast.LENGTH_LONG
                    ).show()
                }
            }
        }
    private val moveToDirLauncher =
        registerForActivityResult(ActivityResultContracts.OpenDocumentTree()) { uri ->
            uri?.let {
                if (copySelectedTo(it)) {
                    runCatching {
                        for (i in mSelected.indices) {
                            if (mSelected[i]) {
                                DocumentsContract.deleteDocument(contentResolver, mImages[i].uri)
                            }
                        }
                    }.onSuccess {
                        Toast.makeText(
                            this,
                            getString(R.string.message_moved, countSelected(), uri.path),
                            Toast.LENGTH_LONG
                        ).show()
                    }.onFailure { e->
                        Log.e("DumbGallery", "Error: removing files")
                        Log.e("DumbGallery", e.stackTrace.toString())
                    }
                } else {
                    Toast.makeText(
                        this,
                        getString(R.string.error_move_failed, uri.path),
                        Toast.LENGTH_LONG
                    ).show()
                }
                openDir()
            }
        }

    private fun copySelectedTo(destDirUri: Uri): Boolean {
        var numCopied = 0
        val destDir = DocumentFile.fromTreeUri(this, destDirUri)
        if (destDir == null) {
            Log.e("DumbGallery", "Creating a DocumentFile for the destDir failed")
        } else {
            for (i in mSelected.indices) {
                if (mSelected[i]) {
                    val destFile = destDir.createFile(mImages[i].mime, mImages[i].name)
                    if (destFile == null) {
                        Log.e("DumbGallery", "Creating a file failed: ${mImages[i].name}")
                    } else {
                        contentResolver.openInputStream(mImages[i].uri)?.buffered()
                            ?.use { inputStream ->
                                contentResolver.openOutputStream(destFile.uri)?.buffered()
                                    ?.use { outputStream ->
                                        if (inputStream.copyTo(outputStream) > 0) { numCopied++ }
                                    }
                            }
                    }
                }
            }
        }

        return (numCopied == countSelected())
    }

    override fun onCreate(savedInstanceState: Bundle?) {
        super.onCreate(savedInstanceState)

        binding = ActivityMainBinding.inflate(layoutInflater)
        setContentView(binding.root)
        binding.fab.setOnClickListener { openDirLauncher.launch(null) }
        binding.gridView.emptyView = binding.textViewEmpty
        binding.gridView.setOnItemClickListener { parent, view, position, _ ->
            if (isSelecting) {
                mSelected[position] = !mSelected[position]
                binding.gridView.adapter.getView(position, view, parent)
                title = getString(R.string.title_select_mode, countSelected())
            } else {
                val intent = Intent(this@MainActivity, DetailsActivity::class.java).apply {
                    putExtra(DetailsActivity.KEY_DIR_URI, mCurrentDirUri)
                    putExtra(DetailsActivity.KEY_SORT_ORDER, mSortOrder)
                    putExtra(DetailsActivity.KEY_SORT_IS_DESCENDING, isDescending)
                    putExtra(DetailsActivity.KEY_CURRENT_NO, position)
                    putExtra(DetailsActivity.KEY_SELECTED, mSelected)
                }
                startDetailsActivity.launch(intent)
            }
        }

        val infoPopupTextView = (layoutInflater.inflate(R.layout.info_popup, null) as TextView).apply {
            setHorizontallyScrolling(true)
            movementMethod = ScrollingMovementMethod.getInstance()
        }
        val infoPopup = PopupWindow(this).apply {
            contentView = infoPopupTextView
            isOutsideTouchable = true
            isFocusable = true
            val dm = resources.displayMetrics
            width = dm.widthPixels *2/3
            height = dm.heightPixels /6
        }
        binding.gridView.setOnItemLongClickListener { _, view, position, _ ->
            val img = mImages[position]
            val (width, height) = img.getImageSize(contentResolver)
            val isRecursive = PreferenceManager.getDefaultSharedPreferences(this)
                .getBoolean(getString(R.string.pref_key_recursive), false)

            infoPopupTextView.text = buildSpannedString {
                append("[${position+1}/${mImages.size}]\n")
                bold {
                    if (isRecursive) {
                        val idx = img.uri.path?.lastIndexOf(title.toString()) ?: 0
                        val file = img.uri.path?.substring(idx)?.replace("$title/", "") ?: ""
                        append(file)
                    } else {
                        append(img.name)
                    }
                }
                append("\n${img.lastModifiedStr()}\n$width x $height pixels\n${img.bytesStr()}")
            }

            infoPopup.showAsDropDown(view, 0, 0)
            val param = (infoPopupTextView.rootView.layoutParams as WindowManager.LayoutParams).apply {
                flags = flags or WindowManager.LayoutParams.FLAG_DIM_BEHIND
                dimAmount = 0.3f
            }
            (getSystemService(Context.WINDOW_SERVICE) as WindowManager).updateViewLayout(infoPopupTextView.rootView, param)
            return@setOnItemLongClickListener true
        }

        val prefs = PreferenceManager.getDefaultSharedPreferences(this)
        val currentDir = prefs.getString(getString(R.string.pref_key_current_dir_uri), "")
        if (!currentDir.isNullOrEmpty() && currentDir != "null") {
            mCurrentDirUri = Uri.parse(currentDir)
            openDir()
        }
    }

    override fun onPause() {
        val prefs = PreferenceManager.getDefaultSharedPreferences(this)
        prefs.edit()
            .putString(getString(R.string.pref_key_current_dir_uri), mCurrentDirUri.toString())
            .putString(getString(R.string.pref_key_current_sort_order), mSortOrder.name)
            .putBoolean(getString(R.string.pref_key_is_descending), isDescending)
            .apply()

        super.onPause()
    }

    override fun onCreateOptionsMenu(menu: Menu): Boolean {
        menuInflater.inflate(R.menu.menu_main, menu)

        return true
    }

    override fun onPrepareOptionsMenu(menu: Menu?): Boolean {
        menu?.let {
            for (i in 0 until menu.size()) {
                val item = menu.getItem(i)
                if (item.itemId == R.id.action_select_done) {
                    item.isVisible = isSelecting
                } else {
                    item.isVisible = !isSelecting
                }
            }
        }

        return super.onPrepareOptionsMenu(menu)
    }

    override fun onOptionsItemSelected(item: MenuItem) : Boolean {
        when (item.itemId) {
            R.id.action_sort -> sortDialog()
//            R.id.action_bookmark -> {
//                val prefs = PreferenceManager.getDefaultSharedPreferences(this)
//                val bookmark = prefs.getString(getString(R.string.pref_key_bookmark), "")?.split("|")?.toMutableList()
//                        ?: mutableListOf()
//                if (mCurrentDirName !in bookmark) {
//                    bookmark.add(mCurrentDirName)
//                    prefs.edit()
//                        .putString(getString(R.string.pref_key_bookmark), bookmark.joinToString(separator = "|"))
//                        .apply()
//                }
//
//            }
            R.id.action_select_mode -> switchSelectMode(true)
            R.id.action_select_done -> switchSelectMode(false)
            R.id.action_remove -> {
                if (mSelected.all {s -> !s}) {
                    Toast.makeText(this, getString(R.string.message_nothing_selected), Toast.LENGTH_SHORT).show()
                    return true
                }

                val dialog = ConfirmationDialogFragment.newInstance(getString(R.string.confirm_remove, countSelected()))
                dialog.setListener(
                    object : ConfirmationDialogFragment.Listener {
                        override fun onPositiveClick() {
                            runCatching {
                                for (i in mSelected.indices) {
                                    if (mSelected[i]) {
                                        DocumentsContract.deleteDocument(contentResolver, mImages[i].uri)
                                    }
                                }
                            }.onSuccess {
                                Toast.makeText(
                                    this@MainActivity,
                                    getString(R.string.message_removed, countSelected()),
                                    Toast.LENGTH_LONG
                                ).show()
                            }.onFailure {
                                Toast.makeText(this@MainActivity, getString(R.string.error_remove_failed), Toast.LENGTH_SHORT).show()
                            }

                            openDir()
                        }
                        override fun onNegativeClick() {}
                    }
                )
                dialog.show(supportFragmentManager, "dialog")
            }
            R.id.action_copy, R.id.action_move -> {
                if (mSelected.all {s -> !s}) {
                    Toast.makeText(this, getString(R.string.message_nothing_selected), Toast.LENGTH_SHORT).show()
                    return true
                }
                if (item.itemId == R.id.action_copy) {
                    copyToDirLauncher.launch(null)
                } else {
                    moveToDirLauncher.launch(null)
                }
            }
            R.id.action_settings -> startActivity(Intent(this, SettingsActivity::class.java))
            else -> return super.onOptionsItemSelected(item)
        }

        return true
    }

    private fun switchSelectMode(isEnabled: Boolean) {
        isSelecting = isEnabled
        title = if (isEnabled) getString(R.string.title_select_mode, countSelected()) else getCurrentDirName()
        val color = if (isEnabled) ColorDrawable(Color.GRAY)
                        else ColorDrawable(resources.getColor(R.color.colorPrimary))
        supportActionBar?.setBackgroundDrawable(color)
        invalidateOptionsMenu()
    }

    private fun openDir() {
        mCurrentDirUri?.let {
            val isRecursive = PreferenceManager.getDefaultSharedPreferences(this)
                .getBoolean(getString(R.string.pref_key_recursive), false)
            val images = ImageItem.imagesInDir(contentResolver, it, isRecursive)
            if (images != null) {
                mImages = images
                mSelected = BooleanArray(images.size) { false }
                binding.gridView.adapter = GridViewAdapter(this@MainActivity, mImages)
                val prefs = PreferenceManager.getDefaultSharedPreferences(this)
                mSortOrder = SortOrder.valueOf(
                    prefs.getString(getString(R.string.pref_key_current_sort_order), "NAME") ?: "NAME"
                )
                isDescending = prefs.getBoolean(getString(R.string.pref_key_is_descending), false)

                mImages.sort(mSortOrder, isDescending)
                mSelected = BooleanArray(mImages.size) {false}
                binding.gridView.adapter = GridViewAdapter(this@MainActivity, mImages)

                title = getCurrentDirName()

                return
            }
        }

        mCurrentDirUri = null
        binding.gridView.adapter = null
    }

    private fun getCurrentDirName(): String {
        val dirUri = DocumentsContract.buildDocumentUriUsingTree(
            mCurrentDirUri,
            DocumentsContract.getTreeDocumentId(mCurrentDirUri)
        )
        contentResolver.query(
            dirUri,
            arrayOf(DocumentsContract.Document.COLUMN_DISPLAY_NAME),
            null,
            null,
            null
        ).use { cursor ->
            if (cursor != null) {
                cursor.moveToFirst()
                val index =
                    cursor.getColumnIndex(DocumentsContract.Document.COLUMN_DISPLAY_NAME)
                return cursor.getString(index)
            }
        }

        return ""
    }

    private fun sortDialog() {
        val selected = when (mSortOrder) {
            SortOrder.NAME -> if (!isDescending) 0 else 1
            SortOrder.DATE -> if (!isDescending) 2 else 3
            SortOrder.RANDOM -> 4
        }

        val dialog = SingleChoiceDialogFragment.newInstance(R.string.menu_sort, selected)
        dialog.setListener(
            object : SingleChoiceDialogFragment.Listener {
                override fun onPositiveClick(result: Int) {
                    when (result) {
                        0 -> { mSortOrder = SortOrder.NAME; isDescending = false }
                        1 -> { mSortOrder = SortOrder.NAME; isDescending = true  }
                        2 -> { mSortOrder = SortOrder.DATE; isDescending = false }
                        3 -> { mSortOrder = SortOrder.DATE; isDescending = true  }
                        4 -> { mSortOrder = SortOrder.RANDOM }
                    }
                    mImages.sort(mSortOrder, isDescending)
                    mSelected = BooleanArray(mImages.size) {false}
                    binding.gridView.adapter = GridViewAdapter(this@MainActivity, mImages)
                }
            }
        )
        dialog.show(supportFragmentManager, "dialog")
    }

    private inner class GridViewAdapter(
        context: Context,
        objects: Array<ImageItem>
    ) : ArrayAdapter<ImageItem>(context, 0, objects) {
        private val mLayoutInflater: LayoutInflater = LayoutInflater.from(context)

        override fun getView(position: Int, convertView: View?, parent: ViewGroup): View {
            val view = (convertView ?: mLayoutInflater.inflate(R.layout.grid_item, parent, false)) as ImageView

            if (mSelected[position]) {
                view.setColorFilter(Color.YELLOW, PorterDuff.Mode.OVERLAY)
            } else {
                view.clearColorFilter()
            }
            getItem(position)?.let {
                Glide.with(this@MainActivity)
                    .load(it.uri)
                    .centerCrop()
                    .into(view)
            }

            return view
        }
    }
}
